# This is about the simplest program to run.
# It tests operations LOAD_CONST and RETURN_VALUE.

# It is great to  it in the beginning when you set up for a new
# Python version and need something

17
